open String
open Preterm

type stp = Base of string | Prop | Ar of (stp * stp)

exception ExpectedTypeError of pretrm * stp * stp

type ctx = (string * stp) list

type trm =
    Name of string * stp
  | False | Imp | Forall of stp | Eq of stp | Choice of stp
  | True | Neg | Or | And | Iff | Exists of stp (*** These are normalized away. ***)
  | DB of int * stp
  | Lam of stp * trm
  | Ap of trm * trm

let oo = Ar(Prop,Prop)
let ooo = Ar(Prop,oo)
let imp x y = Ap(Ap(Imp,x),y)
let preneg x = Ap(Neg,x)
let neg x = imp x False
let normneg m =
  begin
    match m with
    | Ap(Ap(Imp,m1),False) -> m1
    | _ -> neg m
  end
let predisj x y = Ap(Ap(Or,x),y)
let preconj x y = Ap(Ap(And,x),y)
let preiff x y = Ap(Ap(Iff,x),y)
let disj x y = imp (neg x) y
let conj x y = neg (imp x (neg y))
let iff x y = Ap(Ap(Eq(Prop),x),y)
let eq a x y = Ap(Ap(Eq(a),x),y)
let forall a m = Ap(Forall(a),Lam(a,m))
let exists a m = neg (forall a (neg m))
let choice a m = Ap(Choice(a),Lam(a,m))

let existsconst a = let ao = Ar(a,Prop) in Lam(ao,neg (forall a (neg (Ap(DB(1,ao),DB(0,ao))))))

let lpar p = if p then "(" else ""

let rpar p = if p then ")" else ""

let rec stp_str_rec a p =
  match a with
    Base(x) -> x
  | Prop -> "$o"
  | Ar(b,c) -> (lpar p) ^ (stp_str_rec b true) ^ ">" ^ (stp_str_rec c false) ^ (rpar p)

let stp_str a = stp_str_rec a false

let rec stp_contains a = function
    Ar(b, c) -> stp_contains a b || stp_contains a c
  | b -> a = b

let rec church_num_body m =
  match m with
  | DB(0,_) -> true
  | Ap(DB(1,_),m1) -> church_num_body m1
  | _ -> false

let rec church_num_body_val m =
  match m with
  | DB(0,_) -> 0
  | Ap(DB(1,_),m1) -> 1 + (church_num_body_val m1)
  | _ -> raise (GenericSyntaxError("Falsely thought I had the body of a Church numeral. BUG"))

let rec trm_str_rec m p =
  match m with
    Name(x,_) -> x
  | Lam(Ar(a1,a2),DB(0,_)) when (a1 = a2) -> "#1:" ^ (stp_str a1) (*** Church Numeral 1 Printed in a special way. - Chad, Feb 2, 2011 ***)
  | Lam(Ar(a1,a2),Lam(a3,m)) when a1 = a2 && a1 = a3 && church_num_body m -> "#" ^ string_of_int (church_num_body_val m) ^ ":" ^ stp_str a1 (*** Church Numerals Printed in a special way. - Chad, Feb 2, 2011 ***)
  | False -> "False"
  | Imp -> "imp"
  | Forall(a) -> "Pi:" ^ (stp_str a)
  | Eq(a) -> "eq:" ^ (stp_str a)
  | Choice(a) -> "Sepsilon:" ^ (stp_str a)
  | DB(i,_) -> "^" ^ (string_of_int i)
  | Lam(a,m) -> (lpar p) ^ "\\_:" ^ (stp_str a) ^ "." ^ (trm_str_rec m false) ^ (rpar p)
  | Ap(m1,m2) ->
      begin
	match m1 with
	| Lam _ -> (lpar p) ^ (trm_str_rec m1 true) ^ " " ^ (trm_str_rec m2 true) ^ (rpar p)
	| _ -> (lpar p) ^ (trm_str_rec m1 false) ^ " " ^ (trm_str_rec m2 true) ^ (rpar p)
      end
(*** If logic has not been normalized away ***)
  | True -> "True"
  | Neg -> "not"
  | Or -> "or"
  | And -> "and"
  | Iff -> "iff"
  | Exists(_) -> "Sigma"

let trm_str m = trm_str_rec m false

(*** This function assumes m is well-typed. ***)
let rec tpof m =
  match m with
    Name(_,a) -> a
  | False -> Prop
  | Imp -> ooo
  | Forall(a) -> Ar(Ar(a,Prop),Prop)
  | Eq(a) -> Ar(a,Ar(a,Prop))
  | Choice(a) -> Ar(Ar(a,Prop),a)
  | DB(_,a) -> a
  | Lam(a,n) -> Ar(a,tpof n)
  | Ap(f,n) ->
      begin
	match (tpof f) with
	  Ar(a,b) -> b
	| _ -> raise (GenericSyntaxError("Non-function applied: " ^ (trm_str m)))
      end
  | _ -> raise (GenericSyntaxError("Unexpected type case - were logical constants normalized away? " ^ (trm_str m)))

let ueq x y = Ap(Ap(Eq(tpof(x)),x),y)

(*** Call this to check application is well typed ***)
let ap (m,n) =
  match (tpof m,tpof n) with
  | (Ar(a,b),c) when a = c -> Ap(m,n)
  | (a,c) -> raise (GenericSyntaxError("Ill typed application (" ^ (trm_str m) ^ " : " ^ (stp_str a) ^ ") @ (" ^ (trm_str n) ^ " : " ^ (stp_str c) ^ ")"))


(** Prints type m as a Coq-formatted string on the out_channel c  **)
let rec print_stp_coq c m h p =
  match m with
  | Base x ->
      let x = try (Hashtbl.find h x) with Not_found -> failwith("print_stp_coq can't find coqname of "^x) in
      Printf.fprintf c "%s" x
  | Prop ->
      Printf.fprintf c "o"
  | Ar(a,b) ->
      begin
	if p then Printf.fprintf c "(";
	print_stp_coq c a h true;
	Printf.fprintf c " --> ";
	print_stp_coq c b h false;
	if p then Printf.fprintf c ")"
      end

(*
let rec print_stp_isar c m h p =
  match m with
    | Base x ->
	      let x = try (Hashtbl.find h x) with Not_found -> failwith("print_stp_isar can't find coqname of "^x) in
          Printf.fprintf c "%s" x
    | Prop ->
        Printf.fprintf c "o"
    | Ar(a,b) ->
        begin
	        if p then Printf.fprintf c "(";
	        print_stp_isar c a h true;
	        Printf.fprintf c " => ";
	        print_stp_isar c b h false;
	        if p then Printf.fprintf c ")"
        end
*)
let rec print_stp_isar c m p =
  match m with
    | Base x ->
        if x <> "$i" then
          Printf.fprintf c "%s" x
        else
          Printf.fprintf c "i"(*FIXME this may clash with a base type that's really called "i"*)
    | Prop ->
        Printf.fprintf c "o"
    | Ar(a,b) ->
        begin
(*	        if p then*) Printf.fprintf c "(";
	        print_stp_isar c a true;
	        Printf.fprintf c "=>";
	        print_stp_isar c b false;
(*	        if p then*) Printf.fprintf c ")";
	        flush c
        end

let rec print_stp_coq2 c m p =
  match m with
  | Base _ ->
      Printf.fprintf c "set"
  | Prop ->
      Printf.fprintf c "prop"
  | Ar(a,b) ->
      begin
	if p then Printf.fprintf c "(";
	print_stp_coq2 c a true;
	Printf.fprintf c " > ";
	print_stp_coq2 c b false;
	if p then Printf.fprintf c ")"
      end



(*** Shifting, Substitution, Normalization ***)

let rec shift m i j =
  match m with
    DB(k,_) when k < i -> m
  | DB(k,a) -> DB(k + j,a)
  | Lam(a1,m2) -> Lam(a1,shift m2 (i + 1) j)
  | Ap(m1,m2) -> Ap(shift m1 i j,shift m2 i j)
  | _ -> m

let rec subst m i n =
  match m with
    DB(k,_) when k < i -> m
  | DB(k,_) when k = i -> n
  | DB(k,a) -> DB(k - 1,a)
  | Lam(a1,m2) -> Lam(a1,subst m2 (i + 1) (shift n 0 1))
  | Ap(m1,m2) -> Ap(subst m1 i n,subst m2 i n)
  | _ -> m

(*** Simultaneous Substitution ***)
(*** Assumes no dangling deBruijn indices. ***)
let rec simulsubst m s =
  match m with
    DB(k,_) ->
      let n = List.nth s k in
      if (m = n) then m else n
  | Lam(a1,m1) ->
      let n1 = simulsubst m1 ((DB(0,a1))::(List.map (fun n -> shift n 0 1) s)) in
      if (m1 = n1) then m else Lam(a1,n1)
  | Ap(m1,m2) ->
      let n1 = simulsubst m1 s in
      let n2 = simulsubst m2 s in
      if ((m1 = n1) && (m2 = n2)) then m else Ap(n1,n2)
  | _ -> m

let rec namesubst m x n =
  match m with
    Name(y,a) when (x = y) -> n
  | Lam(a1,m1) -> Lam(a1,namesubst m1 x n)
  | Ap(m1,m2) -> Ap(namesubst m1 x n,namesubst m2 x n)
  | _ -> m

let gen_lam_body a m =
  match m with
  | Lam(_,m1) -> m1
  | _ -> Ap(shift m 0 1,DB(0,a))

let rec termNotFreeIn m i =
  match m with
    DB(j,_) when i = j -> false
  | Ap(m1,m2) -> (termNotFreeIn m1 i) && (termNotFreeIn m2 i)
  | Lam(a,m1) -> termNotFreeIn m1 (i + 1)
  | _ -> true

let rec termNotFreeInL m il =
  match m with
    DB(j,_) when (List.mem j il) -> false
  | Ap(m1,m2) -> (termNotFreeInL m1 il) && (termNotFreeInL m2 il)
  | Lam(a,m1) -> termNotFreeInL m1 (List.map (fun i -> i + 1) il)
  | _ -> true

let rec termNoDBs_rec m i =
  match m with
    DB(j,_) -> if (j < i) then true else false
  | Ap(m1,m2) -> (termNoDBs_rec m1 i) && (termNoDBs_rec m2 i)
  | Lam(a,m1) -> termNoDBs_rec m1 (i + 1)
  | _ -> true

let termNoDBs m = termNoDBs_rec m 0

let rec norm1 d m =
  match m with
    Ap(Lam(a,m1),m2) -> (* beta *)
      let (n1,_) = norm1 d m1 in
      let (n2,_) = norm1 d m2 in
      (subst n1 0 n2,true)
  | Lam(_,Ap(m1,DB(0,_))) when (termNotFreeIn m1 0) -> (* eta *)
      (shift m1 0 (- 1),true)
	(*** dneg ***)
  | Ap(Ap(Imp,Ap(Ap(Imp,m1),False)),False) -> (* double negation reduce *)
      let (n1,_) = norm1 d m1 in
      (n1,true)
  | Name(x,_) ->
      begin
	try
	  (Hashtbl.find d x,true)
	with
	| Not_found -> (m,false)
      end
  | Ap(m1,m2) ->
      let (n1,b1) = norm1 d m1 in
      let (n2,b2) = norm1 d m2 in
      if (b1 || b2) then
	(Ap(n1,n2),true)
      else
	(m,false)
  | Lam(a1,m1) ->
      let (n1,b1) = norm1 d m1 in
      if b1 then
	(Lam(a1,n1),true)
      else
	(m,false)
  | _ -> (m,false)

(* beta-eta-delta-dneg *)    
let rec norm d m =
  let (m1,reduced) = norm1 d m in
  if reduced then (norm d m1) else m

let rec betanorm1 d m =
  match m with
    Ap(Lam(a,m1),m2) -> (* beta *)
      let (n1,_) = betanorm1 d m1 in
      let (n2,_) = betanorm1 d m2 in
      (subst n1 0 n2,true)
  | Name(x,_) ->
      begin
	try
	  (Hashtbl.find d x,true)
	with
	| Not_found -> (m,false)
      end
  | Ap(m1,m2) ->
      let (n1,b1) = betanorm1 d m1 in
      let (n2,b2) = betanorm1 d m2 in
      if (b1 || b2) then
	(Ap(n1,n2),true)
      else
	(m,false)
  | Lam(a1,m1) ->
      let (n1,b1) = betanorm1 d m1 in
      if b1 then
	(Lam(a1,n1),true)
      else
	(m,false)
  | _ -> (m,false)
    
(* beta-delta *)    
let rec betanorm d m =
  let (m1,reduced) = betanorm1 d m in
  if reduced then (betanorm d m1) else m

let rec onlybetanorm1 m =
  match m with
    Ap(Lam(a,m1),m2) -> (* onlybeta *)
      let (n1,_) = onlybetanorm1 m1 in
      let (n2,_) = onlybetanorm1 m2 in
      (subst n1 0 n2,true)
  | Ap(m1,m2) ->
      let (n1,b1) = onlybetanorm1 m1 in
      let (n2,b2) = onlybetanorm1 m2 in
      if (b1 || b2) then
	(Ap(n1,n2),true)
      else
	(m,false)
  | Lam(a1,m1) ->
      let (n1,b1) = onlybetanorm1 m1 in
      if b1 then
	(Lam(a1,n1),true)
      else
	(m,false)
  | _ -> (m,false)
    
(* beta (no delta) *)
let rec onlybetanorm m =
  let (m1,reduced) = onlybetanorm1 m in
  if reduced then (onlybetanorm m1) else m

(** Replaces all occurences of 'Neg' by 'implies False' **)
let rec negnorm1 m =
  match m with
  | Ap(Neg,m1) ->
      let (n1,_) = negnorm1 m1 in
      (imp n1 False,true)
  | Neg -> (Lam(Prop,imp (DB(0,Prop)) False),true)
  | Ap(m1,m2) ->
      let (n1,b1) = negnorm1 m1 in
      let (n2,b2) = negnorm1 m2 in
      if (b1 || b2) then
	(Ap(n1,n2),true)
      else
	(m,false)
  | Lam(a1,m1) ->
      let (n1,b1) = negnorm1 m1 in
      if b1 then
	(Lam(a1,n1),true)
      else
	(m,false)
  | _ -> (m,false)

(** applies neg- and betanormalization**)
let onlynegnorm m =
  let (n,_) = negnorm1 m in onlybetanorm n

let rec logicnorm1 m =
  match m with
  | True -> (imp False False,true)
  | Ap(Neg,m1) ->
      let (n1,_) = logicnorm1 m1 in
      (imp n1 False,true)
  | Neg -> (Lam(Prop,imp (DB(0,Prop)) False),true)
  | Ap(Ap(Or,m1),m2) ->
      let (n1,_) = logicnorm1 m1 in
      let (n2,_) = logicnorm1 m2 in
      (disj n1 n2,true)
  | Or -> (Lam(Prop,Lam(Prop,disj (DB(1,Prop)) (DB(0,Prop)))),true)
  | Ap(Ap(And,m1),m2) ->
      let (n1,_) = logicnorm1 m1 in
      let (n2,_) = logicnorm1 m2 in
      (conj n1 n2,true)
  | And -> (Lam(Prop,Lam(Prop,conj (DB(1,Prop)) (DB(0,Prop)))),true)
  | Ap(Ap(Iff,m1),m2) ->
      let (n1,_) = logicnorm1 m1 in
      let (n2,_) = logicnorm1 m2 in
      (iff n1 n2,true)
  | Iff -> (Lam(Prop,Lam(Prop,iff (DB(1,Prop)) (DB(0,Prop)))),true)
  | Ap(Exists(a),Lam(_,m1)) ->
      let (n1,_) = logicnorm1 m1 in
      (exists a n1,true)
  | Exists(a) ->
      let ao = Ar(a,Prop) in
      (Lam(ao,exists a (Ap(DB(1,ao),DB(0,a)))),true)
  | Ap(m1,m2) ->
      let (n1,b1) = logicnorm1 m1 in
      let (n2,b2) = logicnorm1 m2 in
      if (b1 || b2) then
	(Ap(n1,n2),true)
      else
	(m,false)
  | Lam(a1,m1) ->
      let (n1,b1) = logicnorm1 m1 in
      if b1 then
	(Lam(a1,n1),true)
      else
	(m,false)
  | _ -> (m,false)

(* Reduce logical constants to ->, False, forall, =, choice *)
let rec logicnorm m =
  let (n,_) = logicnorm1 m in n

let ideclared = ref false
let basetypestobool = ref false

(*** conversion of pretrm ***)
let rec to_stp (m:pretrm) =
  match m with
    PPropTp -> Prop
  | PIndTp when !basetypestobool -> Prop
  | PName _ when !basetypestobool -> Prop
  | PIndTp ->
      if (!ideclared) then Base("$i")
      else
       begin
        ideclared := true;
        raise DeclareInd
       end
  | PName x ->
     begin
       Base x
     end
  | PAr(a,b) -> Ar(to_stp a,to_stp b)
  | PType -> raise Polymorphism
  | _ -> raise (GenericSyntaxError ((pretrm_str m) ^ " is not a simple type"))

let expected_tp m a b =
  match a with
    Some(aa) ->
      if (aa = b) then () else raise (ExpectedTypeError(m,aa,b))
  | _ -> ()

let rec to_trm (h:(string,(trm * stp) * bool ref) Hashtbl.t) (g:ctx) (m:pretrm) (a:stp option) =
  match m with
  | PAp(PName "Eps",a) -> let b = to_stp a in (Choice b,Ar(Ar(b,Prop),b))
  | PName "In" -> let a = Base "set" in (Name("In",Ar(a,Ar(a,Prop))),Ar(a,Ar(a,Prop)))
  | PName "Subq" -> let a = Base "set" in (Lam(a,Lam(a,Ap(Forall(a),Lam(a,Ap(Ap(Imp,Ap(Ap(Name("In",Ar(a,Ar(a,Prop))),DB(0,a)),DB(2,a))),Ap(Ap(Name("In",Ar(a,Ar(a,Prop))),DB(0,a)),DB(1,a))))))),Ar(a,Ar(a,Prop)))
  | PName x ->
      begin(*** look in g, then in h, then fail ***)
	try
	  to_db m x g a 0
	with
	| Not_found ->
	    try
	      let (zz,o) = Hashtbl.find h x in
	      begin
		(match zz with (_,b) -> expected_tp m a b);
		o := true; (** Mark that it's occurred. - Chad, March 31, 2011 **)
		zz
	      end
	    with
	    | Not_found -> raise (GenericSyntaxError ("Unknown Name " ^ x))
      end
  | PTrue ->
      expected_tp m a Prop;
      (True,Prop)
  | PFalse ->
      expected_tp m a Prop;
      (False,Prop)
  | PNeg -> (Neg,oo)
  | POr ->
      expected_tp m a ooo;
      (Or,ooo)
  | PAnd ->
      expected_tp m a ooo;
      (And,ooo)
  | PIff ->
      expected_tp m a ooo;
      (Iff,ooo)
  | PAp(PAp(PExists,m0),m1) ->
      begin
	expected_tp m a Prop;
	let n0 = to_stp m0 in
	let (n1,a1) = to_trm h g m1 (Some(Ar(n0,Prop))) in
	match a1 with
	  Ar(a1a,Prop) -> (Ap(Exists(a1a),n1),Prop)
	| _ ->
	    raise (ExpectedTypeError(m1,a1,Ar(Base("*"),Prop)))
      end
  | PAp(PAp(PForall,m0),m1) ->
      begin
	expected_tp m a Prop;
	let n0 = to_stp m0 in
	let (n1,a1) = to_trm h g m1 (Some(Ar(n0,Prop))) in
	match a1 with
	  Ar(a1a,Prop) -> (Ap(Forall(a1a),n1),Prop)
	| _ ->
	    raise (ExpectedTypeError(m1,a1,Ar(Base("*"),Prop)))
      end
  | PAp(PExists,m1) ->
      begin
	expected_tp m a Prop;
	let (n1,a1) = to_trm h g m1 None in
	match a1 with
	  Ar(a1a,Prop) -> (Ap(Exists(a1a),n1),Prop)
	| _ ->
	    raise (ExpectedTypeError(m1,a1,Ar(Base("*"),Prop)))
      end
  | PAp(PForall,m1) ->
      begin
	expected_tp m a Prop;
	let (n1,a1) = to_trm h g m1 None in
	match a1 with
	  Ar(a1a,Prop) -> (Ap(Forall(a1a),n1),Prop)
	| _ ->
	    raise (ExpectedTypeError(m1,a1,Ar(Base("*"),Prop)))
      end
  | PAp(PAp(PNIff,m1),m2) ->
      expected_tp m a Prop;
      (preneg (preiff (to_trm_1 h g m1 (Some Prop)) (to_trm_1 h g m2 (Some Prop))),Prop)
  | PNIff ->
      expected_tp m a ooo;
      (Lam(Prop,Lam(Prop,preneg (preiff (DB(1,Prop)) (DB(0,Prop))))),ooo)
  | PImplies ->
      expected_tp m a ooo;
      (Lam(Prop,Lam(Prop,imp (DB(1,Prop)) (DB(0,Prop)))),ooo)
  | PRevImplies ->
      expected_tp m a ooo;
      (Lam(Prop,Lam(Prop,imp (DB(0,Prop)) (DB(1,Prop)))),ooo)
  | PNOr ->
      expected_tp m a ooo;
      (Lam(Prop,Lam(Prop,preneg (predisj (DB(1,Prop)) (DB(0,Prop))))),ooo)
  | PNAnd ->
      expected_tp m a ooo;
      (Lam(Prop,Lam(Prop,preneg (preconj (DB(1,Prop)) (DB(0,Prop))))),ooo)
  | PAp(PAp(PEq,m1),m2) ->
      expected_tp m a Prop;
      let (n1,b1) = to_trm h g m1 None in
      let n2 = to_trm_1 h g m2 (Some b1) in
      ((eq b1 n1 n2),Prop)
  | PEq ->
      begin
	match a with
	  Some(Ar(b1,Ar(b2,Prop)) as aa) when (b1 = b2) ->
	    (Lam(b1,Lam(b1,eq b1 (DB(1,Prop)) (DB(0,Prop)))),aa)
	| Some(aa) ->
	    raise (ExpectedTypeError(m,aa,Ar(Base("*"),Ar(Base("*"),Prop))))
	| None ->
	    raise (GenericSyntaxError ("Cannot determine type of unapplied equality"));
      end
  | PAp(PAp(PNEq,m1),m2) ->
      expected_tp m a Prop;
      let (n1,b1) = to_trm h g m1 None in
      let n2 = to_trm_1 h g m2 (Some b1) in
      (preneg (eq b1 n1 n2),Prop)
  | PNEq ->
      begin
	match a with
	  Some(Ar(b1,Ar(b2,Prop)) as aa) when (b1 = b2) ->
	    (Lam(b1,Lam(b1,preneg (eq b1 (DB(1,Prop)) (DB(0,Prop))))),aa)
	| Some(aa) ->
	    raise (ExpectedTypeError(m,aa,Ar(Base("*"),Ar(Base("*"),Prop))))
	| None ->
	    raise (GenericSyntaxError ("Cannot determine type of unapplied negated equality"));
      end
  | PAp(m1,m2) ->
      begin
	let (n1,a1) = to_trm h g m1 None in
	match a1 with
	  Ar(a1a,a1b) ->
	    expected_tp m a a1b;
	    let n2 = to_trm_1 h g m2 (Some a1a) in
	    (Ap(n1,n2),a1b)
	| _ -> raise (GenericSyntaxError("Non-function applied: " ^ (trm_str n1)))
      end
  | PULam(xl,m1) ->
      begin
	match a with
	| Some(aa) -> to_ulam h g xl m1 aa
	| None -> raise (GenericSyntaxError("Cannot infer type omitted from lambda"))
      end
  | PLam(xl,m1) ->
      to_lam h g xl m1 a
  | PAll(xl,m1) ->
      expected_tp m a Prop;
      (to_all h g xl m1,Prop)
  | PEx(xl,m1) ->
      expected_tp m a Prop;
      (to_exists h g xl m1,Prop)
  | PExU(x,a1,m1) ->
      expected_tp m a Prop;
      begin
	let atp = to_stp a1 in
	match to_trm h ((x,atp)::g) m1 (Some Prop) with
	| (m1a,_) -> (Ap(Lam(Ar(atp,Prop),Ap(Exists(atp),Lam(atp,Ap(Ap(And,Ap(DB(1,Ar(atp,Prop)),DB(0,atp))),Ap(Forall(atp),Lam(atp,Ap(Ap(Imp,Ap(DB(2,Ar(atp,Prop)),DB(0,atp))),Ap(Ap(Eq(atp),DB(0,atp)),DB(1,atp))))))))),Lam(atp,m1a)),Prop)
      end
  | PChoice((x,xa),m1) ->
      let xaa = to_stp xa in
      let n1 = to_trm_1 h ((x,xaa)::g) m1 (Some Prop) in
      (choice xaa n1,xaa)
  | POf(m1,m2) ->
      let b = to_stp m2 in
      expected_tp m1 a b;
      to_trm h g m1 (Some b)
  | PDef(m1,_) ->
      to_trm h g m1 a
  | PLet(x,m1,m2) ->
      begin
	let (m1a,aa) = to_trm h g m1 None in
	let (m2b,bb) = to_trm h ((x,aa)::g) m2 a in
	(Ap(Lam(aa,m2b),m1a),bb)
      end
  | PTLet(x,a1,m1,m2) ->
      begin
	let (m1a,aa) = to_trm h g m1 (Some (to_stp a1)) in
	let (m2b,bb) = to_trm h ((x,aa)::g) m2 a in
	(Ap(Lam(aa,m2b),m1a),bb)
      end
  | _ -> raise (GenericSyntaxError ("Ill-formed term " ^ (pretrm_str m)))
and to_trm_1 h g m a = let (n,_) = to_trm h g m a in n
and to_ulam h (g:ctx) xl m a =
  match xl with
    (x::xr) ->
      begin
	match a with
	  Ar(a1,a2) ->
	    let (n1,_) = to_ulam h ((x,a1)::g) xr m a2 in
	    (Lam(a1,n1),a)
	| _ ->
	    raise (ExpectedTypeError(m,a,Ar(Base("_"),Base("*"))))
      end
  | [] -> to_trm h g m (Some a)
and to_lam h (g:ctx) xl m a =
  match xl with
    ((x,xa)::xr) ->
      begin
	let xaa = to_stp xa in
	match a with
	  Some(Ar(a1,a2) as aa) when (a1 = xaa) ->
	    let (n1,_) = to_lam h ((x,xaa)::g) xr m (Some a2) in
	    (Lam(xaa,n1),aa)
	| Some(aa) ->
	    raise (ExpectedTypeError(m,aa,Ar(xaa,Base("*"))))
	| None ->
	    let (n1,b) = to_lam h ((x,xaa)::g) xr m None in
	    (Lam(xaa,n1),Ar(xaa,b))
      end
  | [] -> to_trm h g m a
and to_all h (g:ctx) xl m =
  match xl with
    ((x,xa)::xr) ->
      begin
	let xaa = to_stp xa in
	let n1 = to_all h ((x,xaa)::g) xr m in
	forall xaa n1
      end
  | [] -> to_trm_1 h g m (Some Prop)
and to_exists h (g:ctx) xl m =
  match xl with
    ((x,xa)::xr) ->
      begin
	let xaa = to_stp xa in
	let n1 = to_exists h ((x,xaa)::g) xr m in
	Ap(Exists(xaa),Lam(xaa,n1))
      end
  | [] ->
      to_trm_1 h g m (Some Prop)
and to_db m x g a i =
  match g with
    ((y,b)::gr) -> 
      if (x = y) then
	begin
	  expected_tp m a b;
	  (DB(i,b),b)
	end
      else
	to_db m x gr a (i + 1)
  | [] -> raise Not_found

let neg_body m =
  match m with
    Ap(Ap(Imp,n),False) -> Some n
  | _ -> None

let eq_body m =
  match m with
      Ap (Ap (Eq a, x), y) -> Some (a, x, y)
    | _ -> None

let neg_p m =
  match (neg_body m) with Some _ -> true | None -> false

(*This is like "head_spine" below, but accepts
  a bound to the number of "Ap"s it is to go
  through before stopping.*)
let bounded_head_spine n m =
  let rec head_spine_aux n m args =
    if n = 0 then (m, args)
    else
      match m with
          Ap (f, a) -> head_spine_aux (n - 1) f (a :: args)
        | _ -> (m, args)
  in
    head_spine_aux n m []

(*Unbounded form of "bounded_head_spine": it
  fully unfolds a term into the head (a non-applicative term)
  and the spine of terms it is applied to.*)
let head_spine m = bounded_head_spine (-1) m

let rec rtp a =
  match a with
    Ar(a1,a2) -> rtp a2
  | _ -> a

let rec argtps_rtp_rec a =
  match a with
  | Ar(a1,a2) ->
      let (al,r) = argtps_rtp_rec a2 in
      (a1::al,r)
  | _ -> ([],a)

let argtps_rtp a = argtps_rtp_rec a


(*List the constants that occur in a term, as name-type pairs.
  NOTE the resulting list doesn't contain duplicate constants.*)
let rec consts_of_trm acc : trm -> ctx = function
    Name (name, stp) ->
    begin
    try
      let stp' = List.assoc name acc
      in
        if stp' <> stp then failwith "Different types for same constant?"
        else acc
    with Not_found ->
      (name, stp) :: acc
    end
  | Lam (_, trm) -> consts_of_trm acc trm
  | Ap (trm1, trm2) ->
    let acc' = consts_of_trm acc trm1
    in consts_of_trm acc' trm2
  | False | Imp | Forall _ | Eq _ | Choice _
  | True | Neg | Or | And | Iff | Exists _
  | DB (_, _) -> acc

(*Produce a list of names (strings) of the base types that occur in a type.
  NOTE the resulting list doesn't contain duplicate names.*)
let rec base_types acc : stp -> string list = function
   Base name ->
   if List.mem name acc then acc
   else name :: acc
 | Prop -> acc
 | Ar (stp1, stp2) ->
   let acc' = base_types acc stp1
   in base_types acc' stp2

(*Lists the base types referenced in a term.
  NOTE the resulting list doesn't contain duplicate elements.*)
let rec base_types_of_trm acc : trm -> string list = function
    Name (name, stp) -> base_types acc stp
  | Lam (stp, trm) ->
    let acc' = base_types acc stp
    in base_types_of_trm acc' trm
  | Ap (trm1, trm2) ->
    let acc' = base_types_of_trm acc trm1
    in base_types_of_trm acc' trm2
  | Forall stp | Eq stp | Choice stp | Exists stp | DB (_, stp) ->
    base_types acc stp
  | False | Imp | True | Neg | Or | And | Iff -> acc

let fresh_prefix = "__"

let make_fresh_name i = fresh_prefix ^ string_of_int i
let is_fresh_name = StringE.starts_with fresh_prefix

let rec map_names f = function
  Name (x,a) -> f x a
| Lam (t,m) -> Lam(t, map_names f m)
| Ap (m,n) -> Ap(map_names f m, map_names f n)
| m -> m

let rec fold_names f acc = function
  Name (x,a) -> f acc x a
| Lam (t,m) -> let (acc', m') = fold_names f acc m in (acc', Lam(t, m'))
| Ap (m,n) ->
    let (acc' , m') = fold_names f acc  m in
    let (acc'', n') = fold_names f acc' n in
    (acc'', Ap(m', n'))
| m -> (acc, m)

let normalize_fresh_uni =
  map_names (fun x a -> Name((if is_fresh_name x then fresh_prefix else x), a))

let normalize_asc_fun p name ((m,i) as acc) x a =
  if p x
  then
  begin
    let (acc', x') =
      try (acc, List.assoc x m)
      with Not_found -> let i' = i+1 in (((x,i')::m,i'), i')
    in (acc', Name(name x', a))
  end
  else (acc, Name (x,a))

let normalize_fresh_asc m =
  fold_names (normalize_asc_fun is_fresh_name make_fresh_name) ([], 0) m

let normalize_asc m =
  fold_names (normalize_asc_fun (fun x -> true) (fun i -> "c" ^ string_of_int i)) ([], 0) m

let contains_fresh_names m = let ((_, i), _) = normalize_fresh_asc m in i > 0

(***
 gives a version with named vars and ~, ->, ! intended to be more readable for the humans.
 partially based on the printer code for Egal
 ***)
let next_var_name a names othernames =
  let basevarnames = ["x";"y";"z";"w";"u";"v"] in
  let predvarnames = ["p";"q";"r"] in
  let funcvarnames = ["f";"g";"h"] in
  let varnames = if rtp a = Prop then predvarnames else match a with Ar(_,_) -> funcvarnames | _ -> basevarnames in
  let numvarnames = List.length varnames in
  let var_name_of j =
    if j < numvarnames then
      List.nth varnames j
    else
      let k = j / numvarnames in
      let m = j mod numvarnames in
      (List.nth varnames m) ^ string_of_int k
  in
  let i = ref 0 in
  let x = ref (var_name_of 0) in
  try
    while true do
      if List.mem !x names || List.mem !x othernames then (incr i; x := var_name_of !i) else raise Exit
    done;
    ""
  with Exit -> !x

let trm_str_nice_paren q (s,p) = if p <= q then s else "(" ^ s ^ ")"

let rec binderp m =
  match m with
  | Lam(_,_) -> true
  | Ap(Forall(_),Lam(_,_)) -> true
  | _ -> false

let rec trm_str_nice_rec m names othernames =
  match m with
    Name(x,_) -> (x,0)
  | Lam(Ar(a1,a2),DB(0,_)) when (a1 = a2) -> ("#1:" ^ (stp_str a1),0) (*** Church Numeral 1 Printed in a special way. - Chad, Feb 2, 2011 ***)
  | Lam(Ar(a1,a2),Lam(a3,m)) when a1 = a2 && a1 = a3 && church_num_body m -> ("#" ^ string_of_int (church_num_body_val m) ^ ":" ^ stp_str a1,0) (*** Church Numerals Printed in a special way. - Chad, Feb 2, 2011 ***)
  | False -> ("False",0)
  | Imp -> ("imp",0)
  | Forall(a) -> ("Pi:" ^ (stp_str a),0)
  | Eq(a) -> ("eq:" ^ (stp_str a),0)
  | Choice(a) -> ("Sepsilon:" ^ (stp_str a),0)
  | DB(i,_) ->
      begin
	try
	  let x = List.nth names i in
	  (x,0)
	with Failure(_) ->
	  ("^" ^ (string_of_int (i - List.length names)),0)
      end
  | Lam(a,m1) ->
      let x = next_var_name a names othernames in
      ("\\" ^ x ^ ":" ^ (stp_str a) ^ "." ^ (trm_str_nice_paren 1010 (trm_str_nice_rec m1 (x::names) othernames)),1)
  | Ap(Ap(Eq(_),m1),m2) ->
      if binderp m1 then (*** to properly use 'binderish' I would need a separate 'layout' type; I'll use binderp instead and hope it's OK ***)
	((trm_str_nice_paren (-1) (trm_str_nice_rec m1 names othernames)) ^ " = " ^ (trm_str_nice_paren 502 (trm_str_nice_rec m2 names othernames)),502)
      else
	((trm_str_nice_paren 502 (trm_str_nice_rec m1 names othernames)) ^ " = " ^ (trm_str_nice_paren 502 (trm_str_nice_rec m2 names othernames)),502)
  | Ap(Ap(Imp,Ap(Ap(Eq(_),m1),m2)),False) ->
      if binderp m1 then (*** to properly use 'binderish' I would need a separate 'layout' type; I'll use binderp instead and hope it's OK ***)
	((trm_str_nice_paren (-1) (trm_str_nice_rec m1 names othernames)) ^ " <> " ^ (trm_str_nice_paren 502 (trm_str_nice_rec m2 names othernames)),502)
      else
	((trm_str_nice_paren 502 (trm_str_nice_rec m1 names othernames)) ^ " <> " ^ (trm_str_nice_paren 502 (trm_str_nice_rec m2 names othernames)),502)
  | Ap(Ap(Imp,m1),False) ->
      ("~" ^ (trm_str_nice_paren 701 (trm_str_nice_rec m1 names othernames)),701)
  | Ap(Ap(Imp,m1),m2) ->
      if binderp m1 then (*** to properly use 'binderish' I would need a separate 'layout' type; I'll use binderp instead and hope it's OK ***)
	((trm_str_nice_paren (-1) (trm_str_nice_rec m1 names othernames)) ^ " -> " ^ (trm_str_nice_paren 801 (trm_str_nice_rec m2 names othernames)),800)
      else
	((trm_str_nice_paren 800 (trm_str_nice_rec m1 names othernames)) ^ " -> " ^ (trm_str_nice_paren 801 (trm_str_nice_rec m2 names othernames)),800)
  | Ap(Forall(a),Lam(_,m1)) ->
      let x = next_var_name a names othernames in
      ("!" ^ x ^ ":" ^ (stp_str a) ^ "." ^ (trm_str_nice_paren 1010 (trm_str_nice_rec m1 (x::names) othernames)),1)
  | Ap(m1,m2) ->
      if binderp m1 then (*** to properly use 'binderish' I would need a separate 'layout' type; I'll use binderp instead and hope it's OK ***)
	((trm_str_nice_paren (-1) (trm_str_nice_rec m1 names othernames)) ^ " " ^ (trm_str_nice_paren 0 (trm_str_nice_rec m2 names othernames)),1)
      else
	((trm_str_nice_paren 1 (trm_str_nice_rec m1 names othernames)) ^ " " ^ (trm_str_nice_paren 0 (trm_str_nice_rec m2 names othernames)),1)
(*** If logic has not been normalized away ***)
  | True -> ("True",0)
  | Neg -> ("not",0)
  | Or -> ("or",0)
  | And -> ("and",0)
  | Iff -> ("iff",0)
  | Exists(_) -> ("Sigma",0)

let trm_str_nice m =
  let nl = List.map (fun (x,_) -> x) (consts_of_trm [] m) in
  let (s,_) = trm_str_nice_rec m [] nl in
  s

(** Jan 2022 Adapted from Cezary Kaliszyk's checking-i.ml from Proofgold Lava **)
(** term annotated with information about
    free vars and hash **)

exception FTrmBd
let dbbound = 180

let bit_30n = [| 1;2;4;8;16;32;64;128;256;512;1024;2048;4096;8192;16384;32768;65536;131072;262144;524288;1048576;2097152;4194304;8388608;16777216;33554432;67108864;134217728;268435456;536870912 |];;

let mk_bit30nl i = Array.to_list (Array.make (i / 30) 0) @ [bit_30n.(i mod 30)];;

let zero_30nl = [];;

let rec or_30nl l1 l2 =
  match l1, l2 with
  | h1 :: t1, h2 :: t2 -> h1 lor h2 :: or_30nl t1 t2
  | h, [] -> h
  | [], h -> h;;

let bit0_0_30nl = function
  | x0 :: _ -> x0 land 1 = 0 | [] -> true;;

let rec bit_0_30nl l j =
  match l with
  | x0 :: t -> if j < 30 then x0 land (1 lsl j) = 0 else bit_0_30nl t (j - 30)
  | [] -> true;;

let rec shift_right_logical_30nl_1 l =
  match l with
  | x0 :: x1 :: t -> (((x1 land 1) lsl 29) lor (x0 lsr 1)) :: shift_right_logical_30nl_1 (x1 :: t)
  | [x0] -> [x0 lsr 1]
  | [] -> [];;

(** term annotated with information about
    free vars and hash **)
type fhashval = int

let ty_no_h=(Hashtbl.create 1000 : (stp, int) Hashtbl.t);;
let no_ty_h=(Hashtbl.create 1000 : (int, stp) Hashtbl.t);;
let ty_num=ref 0;;

let ty_f ty =
  try Hashtbl.find ty_no_h ty
  with Not_found ->
    incr ty_num;
    Hashtbl.add ty_no_h ty !ty_num;
    Hashtbl.add no_ty_h !ty_num ty;
    !ty_num
and ty_f_rev = Hashtbl.find no_ty_h;;

type ftrm =
  | F_DB of int
  | F_Name of string
  | F_Ap of ftrma * ftrma
  | F_Lam of int * ftrma
  | F_Fal
  | F_Imp of ftrma * ftrma
  | F_All of int * ftrma
  | F_Choice of int
  | F_Eq of int * ftrma * ftrma
and ftrma = ftrm * int list * int * fhashval

(* Hashing of random numbers does not make sense, just return these random numbers *)
module HashtblI1Hash =
  struct
    type t = int
    let equal i j = i=j
    let hash i = i
  end
module HashtblI1 = Hashtbl.Make(HashtblI1Hash)

module HashtblI2Hash =
  struct
    type t = int*int
    let equal i j = i=j
    let hash (i,j) = i lxor (j lsl 1) (* To avoid 'x xor x' = 0 for different x *)
  end
module HashtblI2 = Hashtbl.Make(HashtblI2Hash)

module HashtblI3Hash =
  struct
    type t = int*int*int
    let equal i j = i=j
    let hash (i,j,k) = i lxor (j lxor k)
  end
module HashtblI3 = Hashtbl.Make(HashtblI3Hash)

let fhash_used : unit HashtblI1.t = HashtblI1.create 262144 (*1000*)
let mk_name_hash : (string,ftrma) Hashtbl.t = Hashtbl.create 1000
let mk_choice_hash : (int,ftrma) Hashtbl.t = Hashtbl.create 1000

let mk_ap_hash : ftrma HashtblI2.t = HashtblI2.create 131072(*1000*)
let mk_imp_hash : ftrma HashtblI2.t = HashtblI2.create 131072(*1000*)
let mk_eq_hash : ftrma HashtblI3.t = HashtblI3.create 1000
let mk_lam_hash : ftrma HashtblI2.t = HashtblI2.create 8192 (*1000*)
let mk_all_hash : ftrma HashtblI2.t = HashtblI2.create 32768(*1000*)
let mk_ex_hash : ftrma HashtblI2.t = HashtblI2.create 1000
let mk_fal_ref = ref None

let (uhsh   : ftrma HashtblI3.t) = HashtblI3.create 65536(*100*);;
let (shsh   : ftrma HashtblI3.t) = HashtblI3.create 262144(*100*);;

let rec find_random_index () =
  let pos = Random.bits () in
  if HashtblI1.mem fhash_used pos then find_random_index ()
  else begin
    HashtblI1.add fhash_used pos ();
    pos
  end;;

let mk_db_internal i = F_DB(i),mk_bit30nl i,i+1,find_random_index ();;
let mk_db_array = Array.init dbbound mk_db_internal;;
let mk_db_clear () = Array.iteri (fun i _ -> mk_db_array.(i) <- mk_db_internal i) mk_db_array;;

let ftrm_hash_clear () =
  Random.init 1;
  HashtblI1.clear fhash_used;
  Hashtbl.clear mk_name_hash;
  Hashtbl.clear mk_choice_hash;
  HashtblI2.clear mk_ap_hash;
  HashtblI2.clear mk_imp_hash;
  HashtblI3.clear mk_eq_hash;
  HashtblI2.clear mk_lam_hash;
  HashtblI2.clear mk_all_hash;
  HashtblI2.clear mk_ex_hash;
  mk_db_clear ();;

let rec ftrm_str m =
  match m with
  | F_DB(i) -> Printf.sprintf "?%d" i
  | F_Name(x) -> Printf.sprintf "\"%s\"" x
  | F_Ap(m1,m2) -> Printf.sprintf "(Ap %s %s)" (ftrma_str m1) (ftrma_str m2)
  | F_Lam(a,m1) -> Printf.sprintf "(Lam %s)" (ftrma_str m1)
  | F_Fal -> "(Fal)"
  | F_Imp(m1,m2) -> Printf.sprintf "(Imp %s %s)" (ftrma_str m1) (ftrma_str m2)
  | F_All(a,m1) -> Printf.sprintf "(All %s)" (ftrma_str m1)
  | F_Choice(a) -> Printf.sprintf "(Choice %d)" a
  | F_Eq(a,m1,m2) -> Printf.sprintf "(Eq %s %s)" (ftrma_str m1) (ftrma_str m2)
and ftrma_str (m,_,_,_) = ftrm_str m

let mk_db i =
  if i >= dbbound then (Printf.printf "mk_db too high: %d\n%!" i; raise FTrmBd);
  mk_db_array.(i)
;;

let mk_if_not_hashed1 hsh f i =
  try Hashtbl.find hsh i with
  | Not_found ->
     let k = find_random_index () in
     let r = f k i in
     Hashtbl.add hsh i r; r;;

let mk_if_not_ihashed2 hsh f h1 h2 t1 t2 =
  try HashtblI2.find hsh (h1,h2) with
  | Not_found ->
     let k = find_random_index () in
     let r = f k t1 t2 in
     HashtblI2.add hsh (h1,h2) r; r;;

let mk_if_not_ihashed2b hsh f a h1 h2 t1 t2 =
  try HashtblI3.find hsh (a,h1,h2) with
  | Not_found ->
     let k = find_random_index () in
     let r = f k a t1 t2 in
     HashtblI3.add hsh (a,h1,h2) r; r;;

let mk_name_internal k h = F_Name(h),zero_30nl,0,k;;
let mk_name h = mk_if_not_hashed1 mk_name_hash mk_name_internal h;;

let mk_fal () =
  match !mk_fal_ref with
  | None ->
     let r = find_random_index () in
     let ret = (F_Fal,zero_30nl,0,r) in
     mk_fal_ref := Some ret;
     ret
  | Some(ret) -> ret

let mk_imp_internal h ((m1,fv1,cx1,h1) as t1) ((m2,fv2,cx2,h2) as t2) =
  F_Imp(t1,t2),or_30nl fv1 fv2,max cx1 cx2,h;;
let mk_imp ((m1,fv1,cx1,h1) as t1) ((m2,fv2,cx2,h2) as t2) =
  mk_if_not_ihashed2 mk_imp_hash mk_imp_internal h1 h2 t1 t2;;

let mk_ap_internal h ((m1,fv1,cx1,h1) as t1) ((m2,fv2,cx2,h2) as t2) =
  F_Ap(t1,t2),or_30nl fv1 fv2,max cx1 cx2,h;;
let mk_ap ((m1,fv1,cx1,h1) as t1) ((m2,fv2,cx2,h2) as t2) =
  mk_if_not_ihashed2 mk_ap_hash mk_ap_internal h1 h2 t1 t2;;

let mk_eq_internal h a ((m1,fv1,cx1,h1) as t1) ((m2,fv2,cx2,h2) as t2) =
  F_Eq(a,t1,t2),or_30nl fv1 fv2,max cx1 cx2,h;;
let mk_eq a ((m1,fv1,cx1,h1) as t1) ((m2,fv2,cx2,h2) as t2) =
  mk_if_not_ihashed2b mk_eq_hash mk_eq_internal a h1 h2 t1 t2;;

let mk_all_internal h a ((m1,fv1,cx1,h1) as t1) =
  F_All(a,t1),shift_right_logical_30nl_1 fv1,max 0 (cx1 - 1),h;;
let mk_all a ((m1,fv1,cx1,h1) as t1) =
  mk_if_not_ihashed2 mk_all_hash mk_all_internal a h1 a t1;;

let mk_choice_internal h a =
  F_Choice(a),zero_30nl,0,h;;
let mk_choice a =
  mk_if_not_hashed1 mk_choice_hash mk_choice_internal a;;

let mk_lam_internal h a ((m1,fv1,cx1,h1) as t1) =
  F_Lam(a,t1),shift_right_logical_30nl_1 fv1,max 0 (cx1 - 1),h;;
let mk_lam a ((m1,fv1,cx1,h1) as t1) =
  mk_if_not_ihashed2 mk_lam_hash mk_lam_internal (a lsl 10) h1 a t1;;

let uhsh_add j (h,i) t = let tt = h,i,j in HashtblI3.add uhsh tt t;;

type ftask =
  | FT_Lam of int * (int * int)
  | FT_All of int * (int * int)
  | FT_ApR of (ftrma * int * (int * int))
  | FT_ApD of ftrma * (int * int)
  | FT_ImpR of (ftrma * int * (int * int))
  | FT_ImpD of ftrma * (int * int)
  | FT_EqR of int * ftrma * int * (int * int)
  | FT_EqD of int * ftrma * (int * int);;

let rec uptrm_tm_tsk is_tm ((m,fv,cx,h) as t) i j tasks =
  if is_tm then begin
    if cx <= i then uptrm_tm_tsk false t 0 j tasks else
    match HashtblI3.find uhsh (h,i,j) with
    | t' -> uptrm_tm_tsk false t' 0 j tasks
    | exception Not_found ->
      match m with
      | F_Ap(l, r) -> uptrm_tm_tsk true l i j (FT_ApR (r, i, (h, i)) :: tasks)
      | F_Imp(l, r) -> uptrm_tm_tsk true l i j (FT_ImpR (r, i, (h, i)) :: tasks)
      | F_DB k -> uptrm_tm_tsk false (if k < i then t else mk_db (k + j)) 0 j tasks
      | F_All(ty, t') -> uptrm_tm_tsk true t' (i + 1) j (FT_All (ty, (h, i)) :: tasks)
      | F_Lam(ty, t') -> uptrm_tm_tsk true t' (i + 1) j (FT_Lam (ty, (h, i)) :: tasks)
      | F_Eq(ty, l, r) -> uptrm_tm_tsk true l i j (FT_EqR (ty, r, i, (h, i)) :: tasks)
      | _ -> uptrm_tm_tsk false t 0 j tasks (* Does not happen *)
  end else begin
    match tasks with
    | FT_ApR (tmr, i, save) :: tasks -> uptrm_tm_tsk true tmr i j (FT_ApD (t, save) :: tasks)
    | FT_ApD (tml, save) :: tasks -> let ret = mk_ap tml t in uhsh_add j save ret; uptrm_tm_tsk false ret 0 j tasks
    | FT_ImpR (tmr, i, save) :: tasks -> uptrm_tm_tsk true tmr i j (FT_ImpD (t, save) :: tasks)
    | FT_ImpD (tml, save) :: tasks -> let ret = mk_imp tml t in uhsh_add j save ret; uptrm_tm_tsk false ret 0 j tasks
    | FT_All (ty, save) :: tasks -> let ret = mk_all ty t in uhsh_add j save ret; uptrm_tm_tsk false ret 0 j tasks
    | FT_Lam (ty, save) :: tasks -> let ret = mk_lam ty t in uhsh_add j save ret; uptrm_tm_tsk false ret 0 j tasks
    | FT_EqR (ty, tmr, i, save) :: tasks -> uptrm_tm_tsk true tmr i j (FT_EqD (ty, t, save) :: tasks)
    | FT_EqD (ty, tml, save) :: tasks -> let ret = mk_eq ty tml t in uhsh_add j save ret; uptrm_tm_tsk false ret 0 j tasks
    | [] -> t
  end;;

let rec uptrm_f t i j =
  if j = 0 then t else uptrm_tm_tsk true t i j [];;

let mk_lam_norm_internal h a ((m1,fv1,cx1,h1) as t1) =
  match m1 with
  | F_Ap((m2,fv2,cx2,h2) as t1a,(F_DB(j),_,_,_)) when j = 0 && bit0_0_30nl fv2 -> (** eta **)
     uptrm_f t1a 0 (-1)
  | _ -> mk_lam_internal h a t1;;
let mk_lam_norm a ((m1,fv1,cx1,h1) as t1) =
  mk_if_not_ihashed2 mk_lam_hash mk_lam_norm_internal (a lsl 10) h1 a t1;;

let shsh_add sh (h,i) t = let tt = h,i,sh in HashtblI3.add shsh tt t;;

let rec subst_trmtrm_f t j s = (* Non-tail *)
  let rec subst is_tm ((m,fv,cx,h) as t) j ((_,_,_,sh) as s) tasks = (* Tail *)
    if is_tm then begin
      if cx <= j then subst false t j s tasks else (** all free variables below j, so no change at all **)
      if bit_0_30nl fv j then subst false (uptrm_f t j (-1)) j s tasks else (** j is not free in m1, but need to shift those after j down **)
      match HashtblI3.find shsh (h,j,sh) with
      | t' -> subst false t' j s tasks
      | exception Not_found -> match m with
        | F_Ap(l, r) -> subst true l j s (FT_ApR (r, j, (h, j)) :: tasks)
        | F_Imp(l, r) -> subst true l j s (FT_ImpR (r, j, (h, j)) :: tasks)
        | F_DB k -> subst false (uptrm_f s 0 j) j s tasks (* because of FV, j must be equal to k *)
        | F_All(ty, t') -> subst true t' (j + 1) s (FT_All (ty, (h, j)) :: tasks)
        | F_Lam(ty, t') -> subst true t' (j + 1) s (FT_Lam (ty, (h, j)) :: tasks)
        | F_Eq(ty, l, r) -> subst true l j s (FT_EqR (ty, r, j, (h, j)) :: tasks)
        | _ -> subst false t j s tasks
    end else begin
      match tasks with
      | FT_ApR (tmr, j, save) :: tasks -> subst true tmr j s (FT_ApD (t, save) :: tasks)
      | FT_ApD (tml, save) :: tasks ->
         let ((m1,fv1,cx1,h1) as t1) = tml and ((m2,fv2,cx2,h2) as t2) = t in
         let subst_ret =
           try HashtblI2.find mk_ap_hash (h1,h2) 
           with Not_found ->
             let ap_ret =
               match m1 with
               | F_Lam(_,m1b) -> subst_trmtrm_f m1b 0 t2 (* Call to the outside one *)
               | _ -> F_Ap(t1,t2),or_30nl fv1 fv2,max cx1 cx2,find_random_index ()
             in
             HashtblI2.add mk_ap_hash (h1,h2) ap_ret;
             ap_ret
         in
         shsh_add sh save subst_ret;
         subst false subst_ret 0 s tasks
      | FT_ImpR (tmr, j, save) :: tasks -> subst true tmr j s (FT_ImpD (t, save) :: tasks)
      | FT_ImpD (tml, save) :: tasks -> let ret = mk_imp tml t in shsh_add sh save ret; subst false ret 0 s tasks
      | FT_All (ty, save) :: tasks -> let ret = mk_all ty t in shsh_add sh save ret; subst false ret 0 s tasks
      | FT_Lam (ty, save) :: tasks -> let ret = mk_lam_norm ty t in shsh_add sh save ret; subst false ret 0 s tasks
      | FT_EqR (ty, tmr, j, save) :: tasks -> subst true tmr j s (FT_EqD (ty, t, save) :: tasks)
      | FT_EqD (ty, tml, save) :: tasks ->
         let ((m1,fv1,cx1,h1) as t1) = tml and ((m2,fv2,cx2,h2) as t2) = t in
         let subst_ret =
           try HashtblI3.find mk_eq_hash (ty,h1,h2) 
           with Not_found ->
             let eq_ret = F_Eq(ty,t1,t2),or_30nl fv1 fv2,max cx1 cx2,find_random_index () in
             HashtblI3.add mk_eq_hash (ty,h1,h2) eq_ret;
             eq_ret
         in
         shsh_add sh save subst_ret;
         subst false subst_ret 0 s tasks
      | [] -> t
    end in
  subst true t j s [];;

let mk_ap_norm ((m1,fv1,cx1,h1) as t1) ((m2,fv2,cx2,h2) as t2) =
  try HashtblI2.find mk_ap_hash (h1,h2) with
  | Not_found ->
     let r =
       match m1 with
       | F_Lam(_,m1b) ->
          subst_trmtrm_f m1b 0 t2
       | _ -> F_Ap(t1,t2),or_30nl fv1 fv2,max cx1 cx2,find_random_index ()
     in
     HashtblI2.add mk_ap_hash (h1,h2) r; r

let mk_neg t = mk_imp t (mk_fal())

let rec trm_ftrma d m =
  match m with
  | Name(x,_) -> (* drop the type for the ftrm rep *)
     begin
       try
         Hashtbl.find d x
       with Not_found -> mk_name x
     end
  | DB(i,_) -> mk_db i (* drop the type for the ftrm rep *)
  | Lam(a,m1) -> mk_lam_norm (ty_f a) (trm_ftrma d m1)
  | False -> mk_fal()
  | Ap(Ap(Imp,m1),m2) -> mk_imp (trm_ftrma d m1) (trm_ftrma d m2)
  | Ap(Forall(a),m1) -> mk_all (ty_f a) (trm_ftrma d (gen_lam_body a m1))
  | Choice(a) -> mk_choice (ty_f a)
  | Ap(Ap(Eq(a),m1),m2) -> mk_eq (ty_f a) (trm_ftrma d m1) (trm_ftrma d m2)
  | Ap(Ap(Iff,m1),m2) -> mk_eq (ty_f Prop) (trm_ftrma d m1) (trm_ftrma d m2)
  | Imp -> trm_ftrma d (Lam(Prop,Lam(Prop,Ap(Ap(Imp,DB(1,Prop)),DB(1,Prop)))))
  | Forall(a) -> trm_ftrma d (Lam(Ar(a,Prop),Ap(Forall(a),DB(0,Ar(a,Prop)))))
  | Eq(a) -> trm_ftrma d (Lam(a,Lam(a,Ap(Ap(Eq(a),DB(1,a)),DB(0,a)))))
  | Iff -> trm_ftrma d (Lam(Prop,Lam(Prop,Ap(Ap(Eq(Prop),DB(1,Prop)),DB(0,Prop)))))
  | True -> trm_ftrma d (Ap(Ap(Imp,False),False))
  | Ap(Neg,m1) -> trm_ftrma_neg d m1
  | Neg -> trm_ftrma d (Lam(Prop,Ap(Neg,DB(0,Prop))))
  | Ap(Exists(a),Lam(_,m1)) -> trm_ftrma_neg d (Ap(Forall(a),Lam(a,Ap(Neg,m1))))
  | Exists(a) -> trm_ftrma d (Lam(Ar(a,Prop),Ap(Neg,Ap(Forall(a),Lam(a,Ap(Neg,Ap(DB(1,Ar(a,Prop)),DB(0,a))))))))
  | Ap(Ap(Or,m1),m2) -> mk_imp (trm_ftrma_neg d m1) (trm_ftrma d m2)
  | Ap(Ap(And,m1),m2) -> trm_ftrma_neg d (Ap(Ap(Imp,m1),Ap(Ap(Imp,m2),False)))
  | Or -> trm_ftrma d (Lam(Prop,Lam(Prop,Ap(Ap(Imp,Ap(Ap(Imp,DB(1,Prop)),False)),DB(0,Prop)))))
  | And -> trm_ftrma d (Lam(Prop,Lam(Prop,Ap(Ap(Imp,Ap(Ap(Imp,DB(1,Prop)),Ap(Ap(Imp,DB(0,Prop)),False))),False))))
  | Ap(m1,m2) -> mk_ap_norm (trm_ftrma d m1) (trm_ftrma d m2)
and trm_ftrma_neg d m =
  match m with
  | Ap(Neg,m) -> trm_ftrma d m
  | Ap(Ap(Imp,m),False) -> trm_ftrma d m
  | _ ->
     let m1 = trm_ftrma d m in
     mk_neg m1

let ftrma_fhashval (_,_,_,h) = h

let fneg_body (m,_,_,_) =
  match m with
  | F_Imp(t1,(F_Fal,_,_,_)) -> Some(t1)
  | _ -> None

let fneg t =
  match fneg_body t with
  | Some(t1) -> t1
  | None -> mk_neg t

let rec ftrm_trm name_tp cx m =
  match m with
  | F_DB(i) -> DB(i,List.nth cx i)
  | F_Name(x) -> Name(x,Hashtbl.find name_tp x)
  | F_Ap(m1,m2) -> Ap(ftrma_trm name_tp cx m1,ftrma_trm name_tp cx m2)
  | F_Lam(a,m1) -> let a2 = ty_f_rev a in Lam(a2,ftrma_trm name_tp (a2::cx) m1)
  | F_Imp(m1,m2) -> Ap(Ap(Imp,ftrma_trm name_tp cx m1),ftrma_trm name_tp cx m2)
  | F_All(a,m1) -> let a2 = ty_f_rev a in Ap(Forall(a2),Lam(a2,ftrma_trm name_tp cx m1))
  | F_Fal -> False
  | F_Choice(a) -> Choice(ty_f_rev a)
  | F_Eq(a,m1,m2) -> Ap(Ap(Eq(ty_f_rev a),ftrma_trm name_tp cx m1),ftrma_trm name_tp cx m2)
and ftrma_trm name_tp cx (m,_,_,_) = ftrm_trm name_tp cx m

let rec ftrm_trm_2 name_tp cx m =
  match m with
  | F_DB(i) -> DB(i,List.nth cx i)
  | F_Name(x) -> Name(x,ty_f_rev (Hashtbl.find name_tp x))
  | F_Ap(m1,m2) -> Ap(ftrma_trm_2 name_tp cx m1,ftrma_trm_2 name_tp cx m2)
  | F_Lam(a,m1) -> let a2 = ty_f_rev a in Lam(a2,ftrma_trm_2 name_tp (a2::cx) m1)
  | F_Imp(m1,m2) -> Ap(Ap(Imp,ftrma_trm_2 name_tp cx m1),ftrma_trm_2 name_tp cx m2)
  | F_All(a,m1) -> let a2 = ty_f_rev a in Ap(Forall(a2),Lam(a2,ftrma_trm_2 name_tp cx m1))
  | F_Fal -> False
  | F_Choice(a) -> Choice(ty_f_rev a)
  | F_Eq(a,m1,m2) -> Ap(Ap(Eq(ty_f_rev a),ftrma_trm_2 name_tp cx m1),ftrma_trm_2 name_tp cx m2)
and ftrma_trm_2 name_tp cx (m,_,_,_) = ftrm_trm_2 name_tp cx m
                                   
let head_spine_f m =
  let rec head_spine_f_aux ((m,_,_,_) as t) args =
    match m with
      F_Ap(f,a) -> head_spine_f_aux f (a::args)
    | _ -> (t,args)
  in
  head_spine_f_aux m []

type fctx = (string * int) list

(*Lists the base types referenced in a term.
  NOTE the resulting list doesn't contain duplicate elements.*)
let rec base_types_of_trm_f acc : ftrma -> string list = function
    (F_Lam (a, trm),_,_,_) ->
     let acc' = base_types acc (ty_f_rev a) in
     base_types_of_trm_f acc' trm
  | (F_All (a, trm),_,_,_) ->
     let acc' = base_types acc (ty_f_rev a) in
     base_types_of_trm_f acc' trm
  | (F_Ap (trm1, trm2),_,_,_) ->
    let acc' = base_types_of_trm_f acc trm1
    in base_types_of_trm_f acc' trm2
  | (F_Imp (trm1, trm2),_,_,_) ->
    let acc' = base_types_of_trm_f acc trm1
    in base_types_of_trm_f acc' trm2
  | (F_Eq (a, trm1, trm2),_,_,_) ->
    let acc' = base_types_of_trm_f acc trm1
    in base_types_of_trm_f acc' trm2
  | (F_Name (_),_,_,_)
  | (F_DB(_),_,_,_)
  | (F_Fal,_,_,_)
  | (F_Choice(_),_,_,_) -> acc

(*List the constants that occur in a term, as name-type pairs.
  NOTE the resulting list doesn't contain duplicate constants.*)
let rec consts_of_ftrma nh acc : ftrma -> fctx = function
    (F_Name (name),_,_,_) ->
    begin
    try
      ignore (List.assoc name acc);
      acc
    with Not_found ->
      (name, Hashtbl.find nh name) :: acc
    end
  | (F_Lam (_, trm),_,_,_) -> consts_of_ftrma nh acc trm
  | (F_All (_, trm),_,_,_) -> consts_of_ftrma nh acc trm
  | (F_Ap (trm1, trm2),_,_,_) ->
    let acc' = consts_of_ftrma nh acc trm1
    in consts_of_ftrma nh acc' trm2
  | (F_Imp (trm1, trm2),_,_,_) ->
    let acc' = consts_of_ftrma nh acc trm1
    in consts_of_ftrma nh acc' trm2
  | (F_Eq (a, trm1, trm2),_,_,_) ->
    let acc' = consts_of_ftrma nh acc trm1
    in consts_of_ftrma nh acc' trm2
  | (F_DB(_),_,_,_)
  | (F_Fal,_,_,_)
  | (F_Choice(_),_,_,_) -> acc

let rec ftrm_tp nh cx m =
  match m with
  | F_DB(i) -> List.nth cx i
  | F_Name(x) -> Hashtbl.find nh x
  | F_Ap(m1,m2) ->
     begin
       match ftrma_tp nh cx m1 with
       | Ar(a,b) -> check_ftrma_tp nh cx m2 a; b
       | a -> raise (Failure (Printf.sprintf "type mismatch: %s\nexpected: function type\nfound: %s\n" (ftrma_str m1) (stp_str a)))
     end
  | F_Lam(a,m1) -> let a = ty_f_rev a in Ar(a,ftrma_tp nh (a::cx) m1)
  | F_Fal -> Prop
  | F_Imp(m1,m2) -> check_ftrma_tp nh cx m1 Prop; check_ftrma_tp nh cx m2 Prop; Prop
  | F_All(a,m1) -> let a = ty_f_rev a  in check_ftrma_tp nh (a::cx) m1 Prop; Prop
  | F_Choice(a) -> let a = ty_f_rev a in Ar(Ar(a,Prop),a)
  | F_Eq(a,m1,m2) -> let a = ty_f_rev a in check_ftrma_tp nh cx m1 a; check_ftrma_tp nh cx m2 a; Prop
and ftrma_tp nh cx (m,_,_,_) =
  ftrm_tp nh cx m
and check_ftrm_tp nh cx m a =
  let b = ftrm_tp nh cx m in
  if not (a = b) then raise (Failure (Printf.sprintf "type mismatch: %s\nexpected: %s\nfound: %s\n" (ftrm_str m) (stp_str a) (stp_str b)))
and check_ftrma_tp nh cx (m,_,_,_) a =
  check_ftrm_tp nh cx m a
